(*  Title:      FMap.thy
    ID:         $Id: FMap.thy,v 1.1 2007/02/13 12:09:31 hsudhof Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:       Finite maps for Sigma-calculus
                Idea use axiomatic type classes to preserve
                usability of datatype afterwards, i.e. definition
                of an object as a finite map of labels to fields in
                a datatype.
*)

header {* Finite maps with axclasses *}

theory FMap imports ListPre begin

axclass fintype < type

finite_set: " finite (UNIV)"

types ('a, 'b)fmap = "('a :: fintype) ~=> 'b" (infixl "-~>" 50)

axclass inftype < type

infinite: "\<not>finite UNIV"

theorem fset_induct: 
  "P {} ==> (!!x (F :: ('a :: fintype)set). x \<notin> F ==> P F ==> P (insert x F)) ==> P F";
apply (subgoal_tac "finite F")
apply (erule finite_induct);
apply simp+;
apply (rule finite_subset);
thm UNIV_I;
apply (rule subsetI);
apply (rule UNIV_I);
apply (rule finite_set);
done;

theorem fmap_unique: "x = y \<Longrightarrow> (f :: ('a, 'b)fmap) x = f y";
apply (erule ssubst);
by (rule refl);


theorem fmap_case:
  "(F :: ('a -~> 'b)) = empty \<or> (\<exists> x y (F' :: ('a -~> 'b)). F = F'(x \<mapsto> y))";
apply (case_tac "F = empty");
apply (erule disjI1);
apply (rule disjI2);
apply (subgoal_tac "? x. F x ~= None");
apply (erule exE);
apply (case_tac "F x");
apply (rotate_tac 1);
apply (erule notE);
apply assumption;
apply (rule_tac x = x in exI);
apply (rule_tac x = a in exI);
apply (rule_tac x = "F |` (dom F - {x})" in exI);
apply (simp add: restrict_map_def);
apply (rule ext)
apply (case_tac "x = xa");
apply simp;
apply fastsimp;
apply (erule contrapos_np);
apply (subgoal_tac "! x. F x = None");
apply (rule ext);
apply (erule spec);
apply simp;
done;

(* define the witness as a constant function so it may be used in the proof of
the induction scheme below *)
constdefs  
  set_fmap :: "'a -~> 'b \<Rightarrow> ('a * 'b)set"
  "set_fmap F == { (x, y). x : dom F & F x = Some y}"

  pred_set_fmap :: "(('a -~> 'b) \<Rightarrow> bool) \<Rightarrow> (('a * 'b)set) \<Rightarrow> bool"
  "pred_set_fmap P == % S. P (% x. if x : fst ` S then (THE y. (? z. y = Some z & (x, z): S)) else None)" 

  fmap_minus_direct :: "[('a -~> 'b), ('a * 'b)] \<Rightarrow> ('a -~> 'b)" (infixl "--" 50)
  "F -- x == (% z. if (fst x = z & ((F (fst x)) = Some (snd x))) then None else (F z))"

lemma  insert_lem : "insert x A = B \<Longrightarrow> x : B";
by auto;

lemma  set_fmap_pair: "x: set_fmap F \<Longrightarrow> (fst x : dom F & snd x = the (F (fst x)))";
apply (simp add: set_fmap_def);
by auto;


lemma  set_fmap_inv1: "\<lbrakk> fst x: dom F; snd x = the (F (fst x)) \<rbrakk> \<Longrightarrow> (F -- x)(fst x \<mapsto> snd x) = F";
apply (rule ext);
apply (case_tac "xa = fst x");
apply (unfold fmap_minus_direct_def);
apply simp;
apply force;
apply force;
done;

lemma set_fmap_inv2: "fst x ~: dom F \<Longrightarrow> insert x (set_fmap F) = set_fmap (F (fst x \<mapsto> snd x))";
apply (unfold set_fmap_def);
apply (rule equalityI);
apply force;
apply (rule subsetI);
apply (subgoal_tac "(fst xa) \<in> dom (F(fst x \<mapsto> snd x)) \<and> (F(fst x \<mapsto> snd x)) (fst xa) = Some (snd xa)");
apply (erule conjE);
apply (subgoal_tac "xa = x | ((fst xa) : dom F & F (fst xa) = Some (snd xa))");
apply fastsimp;
apply (case_tac "fst x = fst xa");
apply (auto simp: dom_def);
done;


lemma rep_fmap_base: "P (F :: 'a  -~> 'b) = (pred_set_fmap P)(set_fmap F)";
apply (unfold pred_set_fmap_def set_fmap_def);
thm arg_cong;
apply (rule_tac f = P in arg_cong);
apply (rule ext);
apply auto;
apply (rule sym);
thm conjunct1;
apply (rule_tac Q = "a : dom F" in conjunct1); 
apply (rule theI')
thm ex1I;
apply (rule_tac a = "F a" in ex1I);
apply (simp add: dom_def); 
apply simp;
apply (simp add: dom_def); 
apply (case_tac "F x");
apply assumption;
apply auto;
apply (erule notE);
apply (simp add: image_def);
done;

lemma rep_fmap: "(? (Fp :: ('a * 'b)set) (P' :: ('a * 'b)set \<Rightarrow> bool). P (F:: 'a -~> 'b) = P' Fp)";

apply (rule_tac x = "set_fmap F" in exI);
apply (rule_tac x = "pred_set_fmap P" in exI);
by (rule rep_fmap_base);


theorem finite_fsets: "finite (F :: (('a :: fintype) set))";
apply (rule finite_subset);
thm UNIV_I;
apply (rule subsetI);
apply (rule UNIV_I);
apply (rule finite_set);
done;

lemma finite_dom_fmap: "finite ((dom (F :: 'a -~> 'b)) :: (('a :: fintype) set))";
thm finite_subset;
apply (rule_tac B = "UNIV :: (('a :: fintype) set)" in finite_subset);
apply (rule subsetI);
apply (rule UNIV_I);
apply (rule finite_set);
done;

lemma finite_fmap_ran: "finite (ran (F :: ('a :: fintype) -~> 'b))";
apply (unfold ran_def);
apply (subgoal_tac "finite ((% x. the (F x)) ` (dom F))");
apply (simp add: image_def dom_def);
apply (rule_tac P = "(% x. finite x)" in subst);
prefer 2;
apply assumption;
apply force;
apply (rule_tac F = "dom F" in finite_imageI);
apply (rule finite_dom_fmap);
done;

lemma finite_fset_map: "finite (set_fmap (F :: ('a :: fintype) -~> 'b))";
apply (rule_tac B = "((dom F) :: (('a :: fintype) set)) \<times> ran F" in finite_subset);
apply (simp add: set_fmap_def dom_def ran_def);
apply fastsimp;
apply (rule finite_cartesian_product);
apply (rule finite_dom_fmap); 
apply (rule finite_fmap_ran);
done;


lemma rep_fmap_imp: "(! F x z . x \<notin> dom (F :: ('a -~> 'b)) \<longrightarrow> P F \<longrightarrow> P (F (x \<mapsto> z)))
 \<Longrightarrow> (! F x z . x \<notin> fst ` (set_fmap F) \<longrightarrow> (pred_set_fmap P)(set_fmap F) 
      \<longrightarrow> (pred_set_fmap P) (insert (x, z)(set_fmap F)))";

apply clarify;
apply (drule_tac x = F in spec);
apply (drule_tac x = x in spec);
apply (drule_tac x = z in spec);
apply (erule impE);
apply (simp add: set_fmap_def image_def dom_def);
apply (erule impE);
apply (simp add: rep_fmap_base);
apply (subst set_fmap_inv2);
apply (simp add: set_fmap_def image_def dom_def);
apply (simp add: rep_fmap_base);
done;


theorem fmap_induct[rule_format]:
  "P (% x. None) \<longrightarrow> (! (F :: (('a :: fintype) -~> 'b)) x z . x \<notin> dom F \<longrightarrow> P F \<longrightarrow> P (F (x \<mapsto> z)))
  \<longrightarrow> P (F' :: ('a -~> 'b))";

apply (rule_tac t = "P F'" and s = "(pred_set_fmap P)(set_fmap F')" in ssubst);
apply (rule rep_fmap_base);
apply (rule_tac t = "P empty" and s = "(pred_set_fmap P)(set_fmap empty)" in ssubst);
apply (rule rep_fmap_base);
apply clarify;
apply (drule rep_fmap_imp);
apply (subgoal_tac "! F F'. finite F \<longrightarrow> F = (set_fmap F') \<longrightarrow> (pred_set_fmap P) (set_fmap F')");
apply (rotate_tac -1);
apply (drule_tac x = "set_fmap F'" in spec);
apply (rotate_tac -1);
apply (drule_tac x = "F'" in spec);
apply (erule impE);
apply (rule finite_fset_map);
apply (erule impE);
apply (rule refl);
apply (assumption);
apply (rule  allI);
apply (rule allI);
apply (rule impI);
apply (rule_tac x = "F'" in spec);
apply (rotate_tac -1, erule finite_induct);
apply (rule allI);
apply (rename_tac xa F');
apply simp;
apply (subgoal_tac "pred_set_fmap P {}");
apply simp;

apply (unfold set_fmap_def);
apply simp;
apply (subgoal_tac "{(x, y). False} = {}");
apply (erule subst);
apply assumption;
apply simp;
apply (rule allI);
apply (rename_tac xa F');
apply (subgoal_tac "insert x xa = {(x, y). x \<in> dom F' \<and> F' x = Some y} \<longrightarrow> x ~: xa 
               \<longrightarrow> (? F''. (fst x ~: fst ` set_fmap F'') & set_fmap F' = insert x (set_fmap F''))");apply (rule impI);
apply (erule impE);
apply assumption;
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE); 
apply (fold set_fmap_def);
apply (drule_tac x = "F''" in spec);
apply (drule_tac x = "F''" in spec);
apply (drule_tac x = "fst x" in spec);
apply (drule_tac x = "snd x" in spec);
apply (rotate_tac -1);
apply (erule impE);
apply assumption;
apply (rotate_tac -1);
apply (erule impE);
apply simp;
apply (case_tac "xa = set_fmap F''")
apply simp+;
apply (subgoal_tac "! A B x. x ~: A \<longrightarrow> x ~: B \<longrightarrow> insert x A = insert x B \<longrightarrow> A = B");
apply (drule_tac x = xa in spec);
apply (drule_tac x = "set_fmap F''" in spec);
apply (drule_tac x = x in spec);
apply (erule impE);
apply assumption;
apply (erule impE);
apply (subgoal_tac "fst x ~: fst ` set_fmap F'' \<longrightarrow> x ~: set_fmap F''");
apply (erule impE);
apply assumption;
apply simp+;
apply (subgoal_tac "x \<in> set_fmap F'' \<longrightarrow> (fst x \<in> dom F'' \<and> snd x = the (F'' (fst x)))");
prefer 2;
apply (rule impI, rotate_tac -1);
apply (erule set_fmap_pair);
apply (rotate_tac 3)
apply (erule contrapos_np);
apply simp+;
apply (rule allI)+;
apply (rule impI)+;
apply (rule subset_antisym);
apply (rotate_tac -3);
apply (drule_tac B=B in subset_insert);
apply (rotate_tac -1);
apply (erule subst);
apply (erule subst);
apply fastsimp;
apply (rotate_tac -2);
apply (drule_tac B=A in subset_insert);
apply force;
apply force;
apply (rule impI)+;
apply (rule_tac x = "F' -- x" in exI);
apply (rule conjI);
apply (simp add: dom_def fmap_minus_direct_def set_fmap_def);
apply (unfold image_def);
apply simp+;
apply (drule_tac insert_lem);
apply (subgoal_tac "(? y. F' (fst x) = Some y) & F' (fst x) = Some (snd x)");
apply (erule conjE);
apply (rule impI);
apply (rule FalseE);
apply (rotate_tac -2);
apply (erule notE);
apply assumption;
apply fastsimp;
apply (subst set_fmap_inv2);
apply (unfold fmap_minus_direct_def);
apply (subgoal_tac "(fst x, snd x): set_fmap F'");
prefer 2;
apply (rule insert_lem);
apply simp;
apply (subgoal_tac "fst x \<in> dom F' \<and> F' (fst x) = Some (snd x)");
prefer 2;
apply (unfold set_fmap_def);
apply (drule CollectD);
apply fastsimp;
apply (erule conjE);
apply (unfold dom_def);
apply (rule notI);
apply fastsimp;
apply (fold set_fmap_def dom_def fmap_minus_direct_def);
apply (fold set_fmap_def);
apply (frule insert_lem);
apply (drule set_fmap_pair);
apply (erule conjE);
apply (subst set_fmap_inv1);
apply simp+;
done;


end
