(*  Title:      Semantics.thy
    Author:     Ludovic Henrio 
                2008

    Note:      name conventions:
               N component name'
               ci component
		itf interfaces
		i1 interface name
		S: component system consdered
              v: Value
*)

header {* Semantics of GCM Components *}


theory GCMSemantics imports GCM begin



subsection{* Component Behaviour *} 

datatype Component_or_Err = Error | Comp "Component"

definition UpdateFuture:: " Fid\<Rightarrow>Value\<Rightarrow> Value \<Rightarrow>Value" (* replaces Fid by Value1 in Value2*)
where "UpdateFuture f v1 v2 \<equiv> if (f\<in> (snd v2)) then (fst v2,snd v2-{f}\<union>(snd v1))
                      else v2"

inductive red :: "[Component, Component, Component] => bool"  ("_ \<turnstile> _ \<rightarrow>\<^sub>R _" 50) and
          comm:: "[ Component, Name, Fid, Value, Component] => bool"  (" \<turnstile>_ -[_,_,_]\<rightarrow>\<^sub>O _ " 50)
(* S \<turnstile> c1\<rightarrow>\<^sub>R c2 means in the system S, c1 is reduced to c2; 
    \<turnstile> c1 -[i1, i, v]\<rightarrow>\<^sub>O c2 means  c1 is reduced to c2; 
                          emitting request with id i and param v on interface i1*)
  where
   Tau:   "\<lbrakk> (PintState  s,Tau,s2)\<in>(Behaviour s)\<rbrakk> \<Longrightarrow> 
                 S\<turnstile> (Primitive itf s) \<rightarrow>\<^sub>R(Primitive itf (s\<lparr>PintState:=s2\<rparr>))" |

   RcvResult:
    "\<lbrakk>(PintState  s,ReceiveResult f v,s2)\<in>(Behaviour s);f\<in>ComputedRqs S;findResult S f=Some v \<rbrakk> \<Longrightarrow> 
         S\<turnstile> (Primitive itf s) \<rightarrow>\<^sub>R(Primitive itf (s\<lparr>PintState:=s2,
             Pqueue:=map (\<lambda>R. R\<lparr>parameter:=UpdateFuture f v (parameter R)\<rparr>) (Pqueue s),
             PcomputedResults:=map (\<lambda>R. R\<lparr>fValue:=UpdateFuture f v (fValue R)\<rparr>) (PcomputedResults s)\<rparr>))" |
   RcvResultComposite:
    "\<lbrakk>f\<in>ComputedRqs S;findResult S f=Some v \<rbrakk> \<Longrightarrow> 
         S\<turnstile> (Composite itf subCp bindings s) \<rightarrow>\<^sub>R(Composite itf subCp bindings (s\<lparr>
             Cqueue:=map (\<lambda>R. R\<lparr>parameter:=UpdateFuture f v (parameter R)\<rparr>) (Cqueue s),
             CcomputedResults:=map (\<lambda>R. R\<lparr>fValue:=UpdateFuture f v (fValue R)\<rparr>) (CcomputedResults s)\<rparr>))" |

   Call:  "\<lbrakk> (PintState  s,Call i1 v f,s2)\<in>(Behaviour s);f\<notin>(set (RqIdList S))\<rbrakk> \<Longrightarrow> 
                 \<turnstile> (Primitive itf s) -[i1,f,v]\<rightarrow>\<^sub>OPrimitive itf (s\<lparr>PintState:=s2\<rparr>)" |

   EndService:"\<lbrakk>(PintState  s,EndService f v,s2)\<in>(Behaviour s)\<rbrakk> \<Longrightarrow> 
                 S\<turnstile> (Primitive itf s) \<rightarrow>\<^sub>R Primitive itf 
                                   (s\<lparr>PintState:=s2, 
				     PcomputedResults:=(PcomputedResults s) @ [\<lparr>fid=f,fValue=v\<rparr>] \<rparr>)" |

   ServeNext:"\<lbrakk>(Pqueue s)=R#queue;(PintState  s,NewService i v (id R),s2)\<in>(Behaviour s)\<rbrakk> \<Longrightarrow> 
                 S\<turnstile> (Primitive itf s) \<rightarrow>\<^sub>R Primitive itf 
                                   (s\<lparr>PintState:=s2, 
				     Pqueue:=queue \<rparr>)" |

  Hierarchy: "\<lbrakk> N\<in> dom subCp; S\<turnstile> the (subCp N) \<rightarrow>\<^sub>R c2 \<rbrakk> \<Longrightarrow>
	     S\<turnstile> (Composite itf subCp bindings s) \<rightarrow>\<^sub>R Composite itf (subCp(N\<mapsto>c2)) bindings s" |

  CommBrothers:"\<lbrakk> \<lparr>src=N.\<^sub>i i1,dest=N'.\<^sub>i i2\<rparr>\<in>bindings; \<turnstile> the (subCp N) -[i1,f,v]\<rightarrow>\<^sub>O c2 \<rbrakk> \<Longrightarrow> 
                   (* N,N'\<in>subCp is ensured by CorrectComponent Sturcture S*)
	     S\<turnstile> (Composite itf subCp bindings s) \<rightarrow>\<^sub>RComposite itf 
                   (subCp(N\<mapsto>c2)(N'\<mapsto>((the (subCp N'))\<leftarrow>\<lparr>id=f,parameter=v,invokedItf=i2\<rparr>) )) bindings s" |
                                          
  CommParent:"\<lbrakk> \<lparr>src=N.\<^sub>i i1,dest=This i2\<rparr>\<in>bindings; \<turnstile> the (subCp N) -[i1,f,v]\<rightarrow>\<^sub>O c2 \<rbrakk> \<Longrightarrow> 
	     S\<turnstile> (Composite itf subCp bindings s) \<rightarrow>\<^sub>R 
                  (Composite itf (subCp(N\<mapsto>c2)) bindings s)\<leftarrow>\<lparr>id=f,parameter=v,invokedItf=i2\<rparr>" |
                      (* N\<in>subCp is ensured by CorrectComponent Sturcture S*)

  CommChild:"\<lbrakk> Cqueue s= R#Q; \<lparr>src=This (invokedItf R),dest=N'.\<^sub>i i2\<rparr>\<in>bindings; i\<notin>(set (RqIdList S))\<rbrakk> \<Longrightarrow> 
	     S\<turnstile> (Composite itf subCp bindings s) \<rightarrow>\<^sub>R Composite itf 
                   (subCp(N'\<mapsto>((the (subCp N'))\<leftarrow>\<lparr>id=i,parameter=v,invokedItf=i2\<rparr>))) bindings 
                              (s\<lparr>Cqueue:=Q,CcomputedResults:=CcomputedResults s@ [\<lparr>fid=id R,fValue=(0,{i})\<rparr>]\<rparr>)" |
			(* N'\<in>subCp and (invokedItf R)is an interface of composite
                            are ensured by CorrectComponent Sturcture S
                            (kind (the (itf (invokedItf R)))) = Server is ensured by correct component structure S*)

  CompositeCall:"\<lbrakk> Cqueue s= R#Q; (kind (the (itf (invokedItf R)))) = Client; i\<notin>(set (RqIdList S)) \<rbrakk> \<Longrightarrow>
	     \<turnstile> (Composite itf subCp bindings s) -[(invokedItf R),i,(parameter R)]\<rightarrow>\<^sub>O Composite itf subCp bindings 
                               (s\<lparr>Cqueue:=Q,CcomputedResults:=CcomputedResults s@ [\<lparr>fid=id R,fValue=(0,{i})\<rparr>] \<rparr>)"
  (* it is a call + an end service *)
                 (* 0 is strange here! *)
		 (* N'\<in>subCp and (invokedItf R)is an interface of composite
                                     are ensured by CorrectComponent Sturcture S*)


(*!! prove that the semantics maintains the well formedness, ie referenced futures still exist and distinct rqidlist !!! *)

end;