(*  Title:      ListPre.thy
    ID:         $Id: ListPre.thy,v 1.1 2007/02/13 12:09:34 hsudhof Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:       List lemmata and other general stuff as preparation for ASP
*)

header {* List features *}

theory ListPre imports Main begin

lemma drop_lem [rule_format]: "\<lbrakk> drop n obi = drop n f; length obi = length f;  n < length f \<rbrakk>
       \<Longrightarrow> obi ! n  = f ! n ";
thm drop_Suc_conv_tl;
apply (subgoal_tac "obi ! n # drop (Suc n) obi = f ! n # drop (Suc n) f");
apply simp;
apply (simp add: drop_Suc_conv_tl);
done;


lemma nth_mem' [simp]: "n < length xs ==> xs!n mem xs"
by (auto simp add: set_conv_nth mem_iff)


lemma mem_append_lem': "x mem (l @ [y]) \<Longrightarrow> x mem l | (x = y)"
apply (subgoal_tac "x mem (l @ [y]) \<longrightarrow> x mem l | (x = y)");
apply (erule impE);
apply assumption+;
apply (induct_tac l);
apply simp;
apply (rule impI);
apply (case_tac "x = a");
apply simp;
apply simp;
done;

lemma nth_last: "length l = n \<Longrightarrow> (l @ [x])! n = x";
by auto;


lemma take_n: "\<lbrakk>  take n l = take n g; n + 1 <= length g; length l = length g \<rbrakk>
     \<Longrightarrow> take (n + 1)(l [n := g ! n]) = take (n + 1) g";
apply simp;
(*ML "thm \"take_Suc_conv_app_nth\""; *)
(* ?i < length ?xs \<Longrightarrow> take (Suc ?i) ?xs = take ?i ?xs @ [?xs ! ?i]" *)
apply (subst take_Suc_conv_app_nth);
apply simp;
(* ML "thm \"upd_conv_take_nth_drop\""; *)
(* ?i < length ?xs \<Longrightarrow> ?xs[?i := ?a] = take ?i ?xs @ ?a # drop (Suc ?i) ?xs" *)
apply (subst upd_conv_take_nth_drop);
apply simp;
apply (subst take_Suc_conv_app_nth);
apply simp;
apply simp;
done;

lemma drop_n_lem: "\<lbrakk> Suc n <= length l \<rbrakk> \<Longrightarrow> drop (Suc n)(l[n := x]) = drop (Suc n) l";
apply (subst drop_Suc);
apply (subst drop_Suc);
apply (subst upd_conv_take_nth_drop);
apply simp;
apply (subst drop_Suc);
apply (subst drop_tl)+;
apply simp;
done;


lemma drop_n: "\<lbrakk> drop n l = drop n f; n + 1 <= length f; length l = length f \<rbrakk>
     \<Longrightarrow> drop (n + 1)(l [n := g ! n]) = drop (n + 1) f";
apply simp;
apply (subst drop_n_lem);
apply simp;
apply (subst drop_Suc)+;
apply (subst drop_tl)+;
apply simp;
done;


lemma nth_fst: "length l = n + 1 \<Longrightarrow> (l @ [x])! 0 = l ! 0";

apply (subgoal_tac "length l = n + 1 \<longrightarrow> (l @ [x])! 0 = l ! 0");
apply (erule impE);
apply assumption+;
apply (induct_tac l);
apply auto;
done;

lemma nth_zero_app: "\<lbrakk> l ~= []; l ! 0 = x \<rbrakk> \<Longrightarrow> (l @ [y]) ! 0 = x";

apply (subgoal_tac "l ~= [] & l ! 0 = x \<longrightarrow> (l @ [y]) ! 0 = x");
apply (erule impE);
apply (erule conjI);
apply assumption+;

apply (induct_tac l);
by auto;


lemma rev_induct2[consumes 1]: 
"\<lbrakk> length xs = length ys; 
   P [][]; 
  !! x xs y ys. \<lbrakk> length xs = length ys; P xs ys \<rbrakk> \<Longrightarrow> P (xs @ [x])(ys @ [y]) \<rbrakk> \<Longrightarrow> P xs ys";
apply (simplesubst rev_rev_ident[symmetric]);
apply (subgoal_tac "P (rev (rev xs))(rev (rev ys))");
apply simp;
apply (rule_tac xs = "rev xs" and ys = "rev ys" in list_induct2);
apply fastsimp;
apply simp;
apply fastsimp;
done;


(* Similar to induction for 2: see ML "thm \"list_induct2\""; *)

lemma list_induct3: "!! zs ys. \<lbrakk> length xs = length ys; length zs = length xs; P [] [] [];
                !! x xs y ys z zs. \<lbrakk> length xs = length ys; length zs = length xs; P xs ys zs \<rbrakk>
                \<Longrightarrow> P (x # xs)(y # ys)(z # zs) \<rbrakk> \<Longrightarrow> P xs ys zs";

apply(induct xs);
 apply simp
apply(case_tac ys)
 apply simp
apply (case_tac zs);
apply simp;
apply simp;
done;



consts
list_insert :: "('a list) \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a list"
primrec
"list_insert (ah#as) i a = (case i of 
                           0      \<Rightarrow> a#ah#as
                          |Suc j  \<Rightarrow> ah#(list_insert as j a))"
"list_insert [] i a = [a]"


lemma insert_eq [simp]: "\<forall> i \<le>  length e . (list_insert e i a)!i = a "
apply(auto)
apply(induct e)
apply(simp)
apply(case_tac i)
apply(simp)+
done



lemma insert_gt [simp]: " \<forall> i \<le> length e . \<forall> j < i  .  ( (list_insert  e i a)! j = (e!j))"

apply(auto)
apply(induct e)
apply(simp)
apply(case_tac i)
apply(auto)
apply(case_tac j)
apply(auto)
done



lemma insert_lt [simp]: " \<forall> j \<le> length e . \<forall> i \<le>  j  .  ( (list_insert  e i a)! (Suc j) = (e!(j)))"

apply(auto)
apply(induct e)
apply(simp)
apply(case_tac i)
apply(auto)
apply(case_tac j)
apply(auto)
done

lemma insert_lt2 [simp]: " \<forall> j . \<forall> i \<le>  j  .  ( (list_insert  e i a)! (Suc j) = (e!(j)))"

apply(auto)
apply(induct e)
apply(simp)
apply(case_tac i)
apply(auto)
apply(case_tac j)
apply(auto)
done



lemma insert_commute [simp]: "\<forall> i \<le>  length e . (list_insert (list_insert e i b) 0 a) = (list_insert (list_insert  e 0 a ) (Suc i) b)  "
apply(auto)
apply(induct e)
apply(simp)
apply(case_tac i)
apply(simp)+
done

end
